#ifndef CPPROFILER_TESTS_EXECUTION_TEST_HH
#define CPPROFILER_TESTS_EXECUTION_TEST_HH

namespace cpprofiler
{
class Conductor;
}

namespace cpprofiler
{
namespace tests
{
namespace execution
{

void run(Conductor &c);
}
} // namespace tests
} // namespace cpprofiler

#endif