#ifndef CPPROFILER_TESTS_TREE_TEST_HH
#define CPPROFILER_TESTS_TREE_TEST_HH

namespace cpprofiler
{

namespace tests
{

namespace tree_test
{
void run();
}

} // namespace tests

} // namespace cpprofiler

#endif